"""Definitions for the `Energetic` class."""
from mosfit.modules.module import Module


# Important: Only define one ``Module`` class per file.


class Energetic(Module):
    """Template class for energy/velocity conversions."""
